/*
 *  C64.cpp - Put the pieces together
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 */

#include "sysdeps.h"

#include "C64.h"
#include "CPUC64.h"
#include "CPU1541.h"
#include "VIC.h"
#include "SID.h"
#include "CIA.h"
#include "REU.h"
#include "IEC.h"
#include "1541job.h"
#include "Display.h"
#include "Prefs.h"

#if defined(__unix) && !defined(__svgalib__)
#include "CmdPipe.h"
#endif

#ifndef __palmos__
#ifdef FRODO_SC
bool IsFrodoSC = true;
#else
bool IsFrodoSC = false;
#endif
#endif


/*
 *  Constructor: Allocate objects and memory
 */

#ifdef __palmos__
C64::C64(ArmletData *d)
#else
C64::C64()
#endif
{
	int i,j;
	uint8 *p;
	uint32 ui;

#ifdef __palmos__
	armletData=d;
	ui=sizeof(MOS6569State)+sizeof(MOS6581State)+sizeof(MOS6526State)*2+0x8000;
/*
	if (sizeof(MOS6569State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x6569, sizeof(MOS6569State),0);
	if (sizeof(MOS6581State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x6581, sizeof(MOS6581State),0);
	if (sizeof(MOS6526State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x6526, sizeof(MOS6526State),0);
	ShowRequester(REQ_DEBUG, 0x515e, 1, ui,0);
*/		
	ui=0x8000+0x400+sizeof(MOS6510State)+0x800+sizeof(MOS6502State)+sizeof(Job1541State);
/*
	if (sizeof(MOS6510State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x6510, sizeof(MOS6510State),0);
	if (sizeof(MOS6502State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x6502, sizeof(MOS6502State),0);
	if (sizeof(Job1541State) & 3)
		ShowRequester(REQ_DEBUG, 0xbadbad, 0x10b, sizeof(Job1541State),0);
	//ShowRequester(REQ_DEBUG, 0x515e, 2, ui,0);
*/
#endif
	//ShowRequester(REQ_DEBUG, 0xc64, 1,0,0);
	
	// The thread is not yet running
	thread_running = false;
	quit_thyself = false;
	have_a_break = false;

	// System-dependent things
	c64_ctor1();

	// Open display
	//ShowRequester(REQ_DEBUG, 0xc64d159, 1,0,0);
	TheDisplay = new C64Display(this);
#ifdef __palmos__
	if (!TheDisplay) {
		if (!d->errCode) d->errCode=ERR_NEW_DISPLAY;
		return;
	}
	//ShowRequester(REQ_DEBUG, 0xc64d159, 2,0,0);
#endif

	// Allocate RAM/ROM memory
#ifndef __palmos__
	RAM = new uint8[0x10000];
	Basic = new uint8[0x2000];
	Kernal = new uint8[0x2000];
	Char = new uint8[0x1000];
	Color = new uint8[0x0400];
	RAM1541 = new uint8[0x0800];
	ROM1541 = new uint8[0x4000];
#endif

	// Create the chips
#ifdef __palmos__
	//ShowRequester(REQ_DEBUG, 0xc646510, 1,0,0);
	TheCPU = new MOS6510(this, RAM1, RAM2, Basic, Kernal, Char, Color);
	if (!TheCPU) {
		if (!d->errCode) d->errCode=ERR_NEW_CPU;
		return;
	}
	//ShowRequester(REQ_DEBUG, 0xc646510, 2,0,0);
#else
	TheCPU = new MOS6510(this, RAM, Basic, Kernal, Char, Color);
#endif

#ifdef __palmos__
	//ShowRequester(REQ_DEBUG, 0xc641541, 1,0,0);
	TheJob1541 = new Job1541(this, RAM1541);
	if (!TheJob1541) {
		if (!d->errCode) d->errCode=ERR_NEW_JOB1541;
		return;
	//ShowRequester(REQ_DEBUG, 0xc641541, 2,0,0);
	}
#else
	TheJob1541 = new Job1541(RAM1541);
#endif
	//ShowRequester(REQ_DEBUG, 0xc646502, 1,0,0);
	TheCPU1541 = new MOS6502_1541(this, TheJob1541, TheDisplay, RAM1541, ROM1541);
#ifdef __palmos__
	if (!TheCPU1541) {
		if (!d->errCode) d->errCode=ERR_NEW_CPU1541;
		return;
	}
#endif

#ifdef __palmos__
	//ShowRequester(REQ_DEBUG, 0xc646569, 1,0,0);
	TheVIC = TheCPU->TheVIC = new MOS6569(this, TheDisplay, TheCPU, RAM1, RAM2, Char, Color);
	if (!TheVIC) {
		if (!d->errCode) d->errCode=ERR_NEW_VIC;
		return;
	}
#else
	TheVIC = TheCPU->TheVIC = new MOS6569(this, TheDisplay, TheCPU, RAM, Char, Color);
#endif
	//ShowRequester(REQ_DEBUG, 0xc646581, 1,0,0);
	TheSID = TheCPU->TheSID = new MOS6581(this);
#ifdef __palmos__
	if (!TheSID) {
		if (!d->errCode) d->errCode=ERR_NEW_SID;
		return;
	}
#endif
	//ShowRequester(REQ_DEBUG, 0xc6465261, 1,0,0);
	TheCIA1 = TheCPU->TheCIA1 = new MOS6526_1(TheCPU, TheVIC);
#ifdef __palmos__
	if (!TheCIA1) {
		if (!d->errCode) d->errCode=ERR_NEW_CIA1;
		return;
	}
#endif
	//ShowRequester(REQ_DEBUG, 0xc6465262, 1,0,0);
	TheCIA2 = TheCPU->TheCIA2 = TheCPU1541->TheCIA2 = new MOS6526_2(TheCPU, TheVIC, TheCPU1541);
#ifdef __palmos__
	if (!TheCIA2) {
		if (!d->errCode) d->errCode=ERR_NEW_CIA2;
		return;
	}
#endif
#ifdef __palmos__
	//ShowRequester(REQ_DEBUG, 0xc641EC, 1,0,0);
	TheIEC = TheCPU->TheIEC = new IEC(this, TheDisplay);
	if (!TheIEC) {
		if (!d->errCode) d->errCode=ERR_NEW_IEC;
		return;
	}
	//ShowRequester(REQ_DEBUG, 0xc648E0, 1,0,0);
	TheREU = TheCPU->TheREU = new REU(TheCPU, this);
	if (!TheREU) {
		if (!d->errCode) d->errCode=ERR_NEW_REU;
		return;
	}
#else
	TheIEC = TheCPU->TheIEC = new IEC(TheDisplay);
	TheREU = TheCPU->TheREU = new REU(TheCPU);
#endif

	// Initialize RAM with powerup pattern
	//ShowRequester(REQ_DEBUG, 0xc641217, 1,0,0);
#ifdef __palmos__
	for (i=0, p=RAM1; i<256; i++) {
		for (j=0; j<64; j++)
			*p++ = 0;
		for (j=0; j<64; j++)
			*p++ = 0xff;
	}
	for (i=0, p=RAM2; i<256; i++) {
		for (j=0; j<64; j++)
			*p++ = 0;
		for (j=0; j<64; j++)
			*p++ = 0xff;
	}
#else
	for (i=0, p=RAM; i<512; i++) {
		for (j=0; j<64; j++)
			*p++ = 0;
		for (j=0; j<64; j++)
			*p++ = 0xff;
	}
#endif
	//ShowRequester(REQ_DEBUG, 0xc641217, 2,0,0);

	//ShowRequester(REQ_DEBUG, 0xc64c01, 1,0,0);
	// Initialize color RAM with random values
	for (i=0, p=Color; i<1024; i++)
#ifdef __palmos__
	{
		//ShowRequester(REQ_DEBUG, 0xc01, 1, i, 0);
		*p++ = Random() & 0x0f;
	}
#else
		*p++ = rand() & 0x0f;
#endif

	// Clear 1541 RAM
	//ShowRequester(REQ_DEBUG, 0xc64a1541, 1,0,0);
	//ShowRequester(REQ_DEBUG, 0xc641217, 3,0,0);
	memset(RAM1541, 0, 0x800);

	// Open joystick drivers if required
	//ShowRequester(REQ_DEBUG, 0xc64107, 1,0,0);
	//ShowRequester(REQ_DEBUG, 0xc641217, 4,0,0);
	open_close_joysticks(false, false, ThePrefs.Joystick1On, ThePrefs.Joystick2On);
	joykey = 0xff;

#ifdef FRODO_SC
	CycleCounter = 0;
#endif

	// System-dependent things
	c64_ctor2();
	//ShowRequester(REQ_DEBUG, 0xc641217, 5,0,0);
}


/*
 *  Destructor: Delete all objects
 */

C64::~C64()
{
	open_close_joysticks(ThePrefs.Joystick1On, ThePrefs.Joystick2On, false, false);

	delete TheJob1541;
	delete TheREU;
	delete TheIEC;
	delete TheCIA2;
	delete TheCIA1;
	delete TheSID;
	delete TheVIC;
	delete TheCPU1541;
	delete TheCPU;
	delete TheDisplay;

#ifndef __palmos__
	delete[] RAM;
	delete[] Basic;
	delete[] Kernal;
	delete[] Char;
	delete[] Color;
	delete[] RAM1541;
	delete[] ROM1541;
#endif

	c64_dtor();
}


/*
 *  Reset C64
 */

void C64::Reset(void)
{
	TheCPU->AsyncReset();
	TheCPU1541->AsyncReset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheIEC->Reset();
}


/*
 *  NMI C64
 */

void C64::NMI(void)
{
	TheCPU->AsyncNMI();
}


/*
 *  The preferences have changed. prefs is a pointer to the new
 *   preferences, ThePrefs still holds the previous ones.
 *   The emulation must be in the paused state!
 */

void C64::NewPrefs(Prefs *prefs)
{
	//ShowRequester(REQ_DEBUG, 0xfeed, 1, 0,0);
	open_close_joysticks(ThePrefs.Joystick1On, ThePrefs.Joystick2On, prefs->Joystick1On, prefs->Joystick2On);
	//ShowRequester(REQ_DEBUG, 0xfeed, 2, 0,0);
	PatchKernal(prefs->FastReset, prefs->Emul1541Proc);
	//ShowRequester(REQ_DEBUG, 0xfeed, 3, 0,0);

	TheDisplay->NewPrefs(prefs);
	//ShowRequester(REQ_DEBUG, 0xfeed, 4, 0,0);

#ifdef __riscos__
	// Changed order of calls. If 1541 mode hasn't changed the order is insignificant.
	if (prefs->Emul1541Proc) {
		// New prefs have 1541 enabled ==> if old prefs had disabled free drives FIRST
		TheIEC->NewPrefs(prefs);
		TheJob1541->NewPrefs(prefs);
	} else {
		// New prefs has 1541 disabled ==> if old prefs had enabled free job FIRST
		TheJob1541->NewPrefs(prefs);
		TheIEC->NewPrefs(prefs);
	}
#else
	//ShowRequester(REQ_DEBUG, 0xfeed, 5, 0,0);
	TheIEC->NewPrefs(prefs);
	//ShowRequester(REQ_DEBUG, 0xfeed, 6, 0,0);
	TheJob1541->NewPrefs(prefs);
#endif

	//ShowRequester(REQ_DEBUG, 0xfeed, 7, 0,0);
	TheREU->NewPrefs(prefs);
	//ShowRequester(REQ_DEBUG, 0xfeed, 8, 0,0);
	TheSID->NewPrefs(prefs);

	// Reset 1541 processor if turned on
	//ShowRequester(REQ_DEBUG, 0xfeed, 9, 0,0);
	if (!ThePrefs.Emul1541Proc && prefs->Emul1541Proc)
		TheCPU1541->AsyncReset();
	//ShowRequester(REQ_DEBUG, 0xfeed, 10, 0,0);
}


/*
 *  Patch kernal IEC routines
 */

void C64::PatchKernal(bool fast_reset, bool emul_1541_proc)
{
	if (fast_reset) {
		Kernal[0x1d84] = 0xa0;
		Kernal[0x1d85] = 0x00;
	} else {
		Kernal[0x1d84] = orig_kernal_1d84;
		Kernal[0x1d85] = orig_kernal_1d85;
	}

	if (emul_1541_proc) {
		Kernal[0x0d40] = 0x78;
		Kernal[0x0d41] = 0x20;
		Kernal[0x0d23] = 0x78;
		Kernal[0x0d24] = 0x20;
		Kernal[0x0d36] = 0x78;
		Kernal[0x0d37] = 0x20;
		Kernal[0x0e13] = 0x78;
		Kernal[0x0e14] = 0xa9;
		Kernal[0x0def] = 0x78;
		Kernal[0x0df0] = 0x20;
		Kernal[0x0dbe] = 0xad;
		Kernal[0x0dbf] = 0x00;
		Kernal[0x0dcc] = 0x78;
		Kernal[0x0dcd] = 0x20;
		Kernal[0x0e03] = 0x20;
		Kernal[0x0e04] = 0xbe;
	} else {
		Kernal[0x0d40] = 0xf2;	// IECOut
		Kernal[0x0d41] = 0x00;
		Kernal[0x0d23] = 0xf2;	// IECOutATN
		Kernal[0x0d24] = 0x01;
		Kernal[0x0d36] = 0xf2;	// IECOutSec
		Kernal[0x0d37] = 0x02;
		Kernal[0x0e13] = 0xf2;	// IECIn
		Kernal[0x0e14] = 0x03;
		Kernal[0x0def] = 0xf2;	// IECSetATN
		Kernal[0x0df0] = 0x04;
		Kernal[0x0dbe] = 0xf2;	// IECRelATN
		Kernal[0x0dbf] = 0x05;
		Kernal[0x0dcc] = 0xf2;	// IECTurnaround
		Kernal[0x0dcd] = 0x06;
		Kernal[0x0e03] = 0xf2;	// IECRelease
		Kernal[0x0e04] = 0x07;
	}

	// 1541
	ROM1541[0x2ae4] = 0xea;		// Don't check ROM checksum
	ROM1541[0x2ae5] = 0xea;
	ROM1541[0x2ae8] = 0xea;
	ROM1541[0x2ae9] = 0xea;
	ROM1541[0x2c9b] = 0xf2;		// DOS idle loop
	ROM1541[0x2c9c] = 0x00;
	ROM1541[0x3594] = 0x20;		// Write sector
	ROM1541[0x3595] = 0xf2;
	ROM1541[0x3596] = 0xf5;
	ROM1541[0x3597] = 0xf2;
	ROM1541[0x3598] = 0x01;
	ROM1541[0x3b0c] = 0xf2;		// Format track
	ROM1541[0x3b0d] = 0x02;
}


/*
 *  Save RAM contents
 */

#ifndef __palmos__
void C64::SaveRAM(char *filename)
{
	FILE *f;

	if ((f = fopen(filename, "wb")) == NULL)
		ShowRequester("RAM save failed.", "OK", NULL);
	else {
		fwrite((void*)RAM, 1, 0x10000, f);
		fwrite((void*)Color, 1, 0x400, f);
		if (ThePrefs.Emul1541Proc)
			fwrite((void*)RAM1541, 1, 0x800, f);
		fclose(f);
	}
}
#endif


/*
 *  Save CPU state to snapshot
 *
 *  0: Error
 *  1: OK
 *  -1: Instruction not completed
 */

#ifdef __palmos__
uint16 C64::SaveCPUState(uint8 *p1, uint8 *p2)
{
	MOS6510State *state = (MOS6510State *)(p2+0x8000+0x400);
	//ShowRequester(REQ_DEBUG, 0xc0ffee, (uint32)p1, (uint32)p2, (uint32)state);
	TheCPU->GetState(state);
	//ShowRequester(REQ_DEBUG, 0xb00b, 2, ((uint8 *)&(state->pc))-p2, 0);

	if (!state->instruction_complete)
		return 0;

	//ShowRequester(REQ_DEBUG, 0xc0ffee, (uint32)p1, (uint32)p2, (uint32)p1);
	memcpy(p1, RAM1, 0x8000);
	//ShowRequester(REQ_DEBUG, 0xc0ffee, (uint32)p1, (uint32)p2, (uint32)p2);
	memcpy(p2, RAM2, 0x8000);
	//ShowRequester(REQ_DEBUG, 0xc0ffee, (uint32)p1, (uint32)p2, (uint32)(p2+0x8000));
	memcpy(p2+0x8000, Color, 0x400);

	// p1 now contains 32k of memory
	// p2 contains other 32k, then Color and CPU state

	return 0x8000+0x400+sizeof(MOS6510State);
}
#else
int C64::SaveCPUState(FILE *f)
{
	MOS6510State state;
	TheCPU->GetState(&state);

	if (!state.instruction_complete)
		return -1;

	int i = fwrite(RAM, 0x10000, 1, f);
	i += fwrite(Color, 0x400, 1, f);
	i += fwrite((void*)&state, sizeof(state), 1, f);

	return i == 3;
}
#endif


/*
 *  Load CPU state from snapshot
 */

#ifdef __palmos__
uint16 C64::LoadCPUState(uint8 *p1, uint8 *p2)
{
	MOS6510State *state = (MOS6510State *)(p2+0x8000+0x400);

	
	memcpy(RAM1, p1, 0x8000);
	memcpy(RAM2, p2, 0x8000);
	memcpy(Color, p2+0x8000, 0x400);
	TheCPU->SetState(state);

	return 0x8000+0x400+sizeof(MOS6510State);
}
#else
bool C64::LoadCPUState(FILE *f)
{
	MOS6510State state;

	int i = fread(RAM, 0x10000, 1, f);
	i += fread(Color, 0x400, 1, f);
	i += fread((void*)&state, sizeof(state), 1, f);

	if (i == 3) {
		TheCPU->SetState(&state);
		return true;
	} else
		return false;
}
#endif


/*
 *  Save 1541 state to snapshot
 *
 *  0: Error
 *  1: OK
 *  -1: Instruction not completed
 */

#ifdef __palmos__
uint16 C64::Save1541State(uint8 *p)
{
	MOS6502State *state =  (MOS6502State *) (p+0x800);
	TheCPU1541->GetState(state);

	if (!state->idle && !state->instruction_complete)
		return 0;

	memcpy(p, RAM1541, 0x800);

	return 0x800+sizeof(MOS6502State);
}
#else
int C64::Save1541State(FILE *f)
{
	MOS6502State state;
	TheCPU1541->GetState(&state);

	if (!state.idle && !state.instruction_complete)
		return -1;

	int i = fwrite(RAM1541, 0x800, 1, f);
	i += fwrite((void*)&state, sizeof(state), 1, f);

	return i == 2;
}
#endif


/*
 *  Load 1541 state from snapshot
 */

#ifdef __palmos__
uint16 C64::Load1541State(uint8 *p)
{
	MOS6502State *state = (MOS6502State *) (p+0x800);

	memcpy(RAM1541, p, 0x800);
	TheCPU1541->SetState(state);

	return 0x800+sizeof(MOS6502State);
}
#else
bool C64::Load1541State(FILE *f)
{
	MOS6502State state;

	int i = fread(RAM1541, 0x800, 1, f);
	i += fread((void*)&state, sizeof(state), 1, f);

	if (i == 2) {
		TheCPU1541->SetState(&state);
		return true;
	} else
		return false;
}
#endif


/*
 *  Save VIC state to snapshot
 */

#ifdef __palmos__
uint16 C64::SaveVICState(uint8 *p)
{
	MOS6569State *state = (MOS6569State *) p;
	TheVIC->GetState(state);
	return sizeof(MOS6569State);
}
#else
bool C64::SaveVICState(FILE *f)
{
	MOS6569State state;
	TheVIC->GetState(&state);
	return fwrite((void*)&state, sizeof(state), 1, f) == 1;
}
#endif


/*
 *  Load VIC state from snapshot
 */

#ifdef __palmos__
uint16 C64::LoadVICState(uint8 *p)
{
	MOS6569State *state = (MOS6569State *) p;

	TheVIC->SetState(state);
	return sizeof(MOS6569State);
}
#else
bool C64::LoadVICState(FILE *f)
{
	MOS6569State state;

	if (fread((void*)&state, sizeof(state), 1, f) == 1) {
		TheVIC->SetState(&state);
		return true;
	} else
		return false;
}
#endif

/*
 *  Save SID state to snapshot
 */

#ifdef __palmos__
uint16 C64::SaveSIDState(uint8 *p)
{
	MOS6581State *state = (MOS6581State *) p;
	TheSID->GetState(state);
	return sizeof(MOS6581State);
}
#else
bool C64::SaveSIDState(FILE *f)
{
	MOS6581State state;
	TheSID->GetState(&state);
	return fwrite((void*)&state, sizeof(state), 1, f) == 1;
}
#endif


/*
 *  Load SID state from snapshot
 */

#ifdef __palmos__
uint16 C64::LoadSIDState(uint8 *p)
{
	MOS6581State *state = (MOS6581State *) p;

	TheSID->SetState(state);
	return sizeof(MOS6581State);
}
#else
bool C64::LoadSIDState(FILE *f)
{
	MOS6581State state;

	if (fread((void*)&state, sizeof(state), 1, f) == 1) {
		TheSID->SetState(&state);
		return true;
	} else
		return false;
}
#endif


/*
 *  Save CIA states to snapshot
 */

#ifdef __palmos__
uint16 C64::SaveCIAState(uint8 *p)
{
	MOS6526State *state = (MOS6526State *) p;
	TheCIA1->GetState(state);
	state = (MOS6526State *) p+sizeof(MOS6526State);
	TheCIA2->GetState(state);

	return sizeof(MOS6526State)*2;
}
#else
bool C64::SaveCIAState(FILE *f)
{
	MOS6526State state;
	TheCIA1->GetState(&state);

	if (fwrite((void*)&state, sizeof(state), 1, f) == 1) {
		TheCIA2->GetState(&state);
		return fwrite((void*)&state, sizeof(state), 1, f) == 1;
	} else
		return false;
}
#endif


/*
 *  Load CIA states from snapshot
 */

#ifdef __palmos__
uint16 C64::LoadCIAState(uint8 *p)
{
	MOS6526State *state = (MOS6526State *) p;

	TheCIA1->SetState(state);
	state = (MOS6526State *) p+sizeof(MOS6526State);
	TheCIA2->SetState(state);

	return sizeof(MOS6526State)*2;
}
#else
bool C64::LoadCIAState(FILE *f)
{
	MOS6526State state;

	if (fread((void*)&state, sizeof(state), 1, f) == 1) {
		TheCIA1->SetState(&state);
		if (fread((void*)&state, sizeof(state), 1, f) == 1) {
			TheCIA2->SetState(&state);
			return true;
		} else
			return false;
	} else
		return false;
}
#endif


/*
 *  Save 1541 GCR state to snapshot
 */

#ifdef __palmos__
uint16 C64::Save1541JobState(uint8 *p)
{
	Job1541State *state = (Job1541State *) p;
	TheJob1541->GetState(state);
	return sizeof(Job1541State);
}
#else
bool C64::Save1541JobState(FILE *f)
{
	Job1541State state;
	TheJob1541->GetState(&state);
	return fwrite((void*)&state, sizeof(state), 1, f) == 1;
}
#endif


/*
 *  Load 1541 GCR state from snapshot
 */

#ifdef __palmos__
uint16 C64::Load1541JobState(uint8 *p)
{
	Job1541State *state = (Job1541State *) p;

	TheJob1541->SetState(state);
	return sizeof(Job1541State);
}
#else
bool C64::Load1541JobState(FILE *f)
{
	Job1541State state;

	if (fread((void*)&state, sizeof(state), 1, f) == 1) {
		TheJob1541->SetState(&state);
		return true;
	} else
		return false;
}
#endif

#ifndef __palmos__
#define SNAPSHOT_HEADER "FrodoSnapshot"
#endif
#define SNAPSHOT_1541 1

#define ADVANCE_CYCLES	\
	TheVIC->EmulateCycle(); \
	TheCIA1->EmulateCycle(); \
	TheCIA2->EmulateCycle(); \
	TheCPU->EmulateCycle(); \
	if (ThePrefs.Emul1541Proc) { \
		TheCPU1541->CountVIATimers(1); \
		if (!TheCPU1541->Idle) \
			TheCPU1541->EmulateCycle(); \
	}


/*
 *  Save snapshot (emulation must be paused and in VBlank)
 *
 *  To be able to use SC snapshots with SL, SC snapshots are made thus that no
 *  partially dealt with instructions are saved. Instead all devices are advanced
 *  cycle by cycle until the current instruction has been finished. The number of
 *  cycles this takes is saved in the snapshot and will be reconstructed if the
 *  snapshot is loaded into FrodoSC again.
 */
 

#ifdef __palmos__
void C64::SaveSnapshot(uint8 *block1, uint8 *block2)
{
	uint8 *p1, *p2;
	uint8 flags;
	uint8 delay;
	int stat;

	p1=block1;
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p1,(uint32)p1-(uint32)block1,1);
	// SNAPSHOT_HEADER
	*p1++='F'; *p1++='r'; *p1++='o'; *p1++='d'; *p1++='o'; *p1++='S'; 
	*p1++='n'; *p1++='a'; *p1++='p'; *p1++='s'; *p1++='h'; *p1++='o'; 
	*p1++='t'; *p1++='\n';
	*p1++=0;	// Version number 0
	flags = 0;
	if (ThePrefs.Emul1541Proc)
		flags |= SNAPSHOT_1541;
	*p1++=flags;
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p1,(uint32)p1-(uint32)block1,2);
	p1+=SaveVICState(p1);
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p1,(uint32)p1-(uint32)block1,3);
	p1+=SaveSIDState(p1);
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p1,(uint32)p1-(uint32)block1,4);
	p1+=SaveCIAState(p1);
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p1,(uint32)p1-(uint32)block1,5);

	p2=block2;

#ifdef FRODO_SC
	delay = 0;
	do {
		if (!(stat = SaveCPUState(p1, p2))) {	// 0 -> Instruction not finished yet
			ADVANCE_CYCLES;	// Advance everything by one cycle
			delay++;
		}
	} while (!stat);
	p2+=stat;
	*p2++=delay;	// Number of cycles the saved CPUC64 lags behind the previous chips
#else
	//ShowRequester(REQ_DEBUG, 0xb00b, 0, p1-block1,0);
	//ShowRequester(REQ_DEBUG, 0xb00b, 1, p1+0x8000 - block1,0);
	p2+=SaveCPUState(p1, p2);
	//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p2,(uint32)p2-(uint32)block2,1);
	*p2++=0;	// No delay
#endif

	//ShowRequester(REQ_DEBUG, 0xb00b, 3, p2-block2, 0);
	
	if (ThePrefs.Emul1541Proc) {
		memcpy(p2, ThePrefs.DrivePath[0], 256);
		p2+=256;
		//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p2,(uint32)p2-(uint32)block2,2);
#ifdef FRODO_SC
		delay = 0;
		do {
			if (!(stat = Save1541State(p2))) {
				ADVANCE_CYCLES;
				delay++;
			}
		} while (!stat);
		p2+=stat;
		*p2++=delay;
#else
		p2+=Save1541State(p2);
		*p2++=0;	// No delay
		//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p2,(uint32)p2-(uint32)block2,3);
#endif
		p2+=Save1541JobState(p2);
		//ShowRequester(REQ_DEBUG, 0xbabe, (uint32)p2,(uint32)p2-(uint32)block2,4);
	}
}
#else
void C64::SaveSnapshot(char *filename)
{
	FILE *f;
	uint8 flags;
	uint8 delay;
	int stat;

	if ((f = fopen(filename, "wb")) == NULL) {
		ShowRequester("Unable to open snapshot file", "OK", NULL);
		return;
	}

	fprintf(f, "%s%c", SNAPSHOT_HEADER, 10);
	fputc(0, f);	// Version number 0
	flags = 0;
	if (ThePrefs.Emul1541Proc)
		flags |= SNAPSHOT_1541;
	fputc(flags, f);
	SaveVICState(f);
	SaveSIDState(f);
	SaveCIAState(f);

#ifdef FRODO_SC
	delay = 0;
	do {
		if ((stat = SaveCPUState(f)) == -1) {	// -1 -> Instruction not finished yet
			ADVANCE_CYCLES;	// Advance everything by one cycle
			delay++;
		}
	} while (stat == -1);
	fputc(delay, f);	// Number of cycles the saved CPUC64 lags behind the previous chips
#else
	SaveCPUState(f);
	fputc(0, f);		// No delay
#endif

	if (ThePrefs.Emul1541Proc) {
		fwrite(ThePrefs.DrivePath[0], 256, 1, f);
#ifdef FRODO_SC
		delay = 0;
		do {
			if ((stat = Save1541State(f)) == -1) {
				ADVANCE_CYCLES;
				delay++;
			}
		} while (stat == -1);
		fputc(delay, f);
#else
		Save1541State(f);
		fputc(0, f);	// No delay
#endif
		Save1541JobState(f);
	}
	fclose(f);

#ifdef __riscos__
	TheWIMP->SnapshotSaved(true);
#endif
}
#endif


/*
 *  Load snapshot (emulation must be paused and in VBlank)
 */

#ifdef __palmos__
bool C64::LoadSnapshot(uint8 *block1, uint8 *block2)
{
	uint8 *p1, *p2;
	unsigned char Header[14];
	uint8 flags, delay;
	bool error = false;
#ifndef FRODO_SC
	uint8 *vicptr;	// Pointer to VIC data
#endif

	//ShowRequester(REQ_DEBUG, 0xb0bba, (uint32)block1, 0, 1);
	p1=Header;
	*p1++='F'; *p1++='r'; *p1++='o'; *p1++='d'; *p1++='o'; *p1++='S'; 
	*p1++='n'; *p1++='a'; *p1++='p'; *p1++='s'; *p1++='h'; *p1++='o';
	*p1++='t'; *p1++='\n';

	if (memcmp(Header, block1, sizeof(Header))) {
		memcpy(Header, block1, sizeof(Header)-1);
		Header[sizeof(Header)-1]='\0';
		ShowRequester(REQ_BAD_SNAPSHOT, (uint32)Header,0,0,0);
		return false;
	}
	p1=block1+sizeof(Header);

	if (*p1 != 0) {
		ShowRequester(REQ_BAD_SNAPSHOT_VERSION, *p1,0,0,0);
		return false;
	}
	p1++;
	flags = *p1++;
#ifndef FRODO_SC
	vicptr = p1;
#endif

	//ShowRequester(REQ_DEBUG, 0xb0bba, (uint32)p1, (uint32)p1-(uint32)(block1), 2);
	p1+=LoadVICState(p1);
	//ShowRequester(REQ_DEBUG, 0xb0bba, (uint32)p1, (uint32)p1-(uint32)(block1), 3);
	p1+=LoadSIDState(p1);
	//ShowRequester(REQ_DEBUG, 0xb0bba, (uint32)p1, (uint32)p1-(uint32)(block1), 4);
	p1+=LoadCIAState(p1);
	//ShowRequester(REQ_DEBUG, 0xb0bba, (uint32)p1, (uint32)p1-(uint32)(block1), 5);
	p2=block2;
	p2+=LoadCPUState(p1, p2);
	//ShowRequester(REQ_DEBUG, 0xfe77, (uint32)p2, (uint32)p2-(uint32)(block2), 1);

	delay = *p2++;	// Number of cycles the 6510 is ahead of the previous chips
#ifdef FRODO_SC
	// Make the other chips "catch up" with the 6510
	for (i=0; i<delay; i++) {
		TheVIC->EmulateCycle();
		TheCIA1->EmulateCycle();
		TheCIA2->EmulateCycle();
	}
#endif
	if ((flags & SNAPSHOT_1541) != 0) {
		// First switch on emulation
		TheNewPrefs = ThePrefs;
		memcpy(TheNewPrefs.DrivePath[0], p2, 256);
		p2+=256;
		TheNewPrefs.Emul1541Proc = true;
		NewPrefs(&TheNewPrefs);
		ThePrefs = TheNewPrefs;
	
		// Then read the context
		p2+=Load1541State(p2);
		//ShowRequester(REQ_DEBUG, 0xfe77, (uint32)p2, (uint32)p2-(uint32)(block2), 2);
	
		delay = *p2++;	// Number of cycles the 6502 is ahead of the previous chips
#ifdef FRODO_SC
		// Make the other chips "catch up" with the 6502
		for (i=0; i<delay; i++) {
			TheVIC->EmulateCycle();
			TheCIA1->EmulateCycle();
			TheCIA2->EmulateCycle();
			TheCPU->EmulateCycle();
		}
#endif
		p2+=Load1541JobState(p2);
		//ShowRequester(REQ_DEBUG, 0xfe77, (uint32)p2, (uint32)p2-(uint32)(block2), 3);
	} else if (ThePrefs.Emul1541Proc) {	// No emulation in snapshot, but currently active?
		Prefs *prefs = new Prefs(ThePrefs);
		prefs->Emul1541Proc = false;
		NewPrefs(prefs);
		ThePrefs = *prefs;
		delete prefs;
	}

#ifndef FRODO_SC
	p1=vicptr;
	p1+=LoadVICState(p1);	// Load VIC data twice in SL (is REALLY necessary sometimes!)
#endif
	return true;
}
#else
bool C64::LoadSnapshot(char *filename)
{
	FILE *f;

	if ((f = fopen(filename, "rb")) != NULL) {
		char Header[] = SNAPSHOT_HEADER;
		char *b = Header, c = 0;
		uint8 delay, i;

		// For some reason memcmp()/strcmp() and so forth utterly fail here.
		while (*b > 32) {
			if ((c = fgetc(f)) != *b++) {
				b = NULL;
				break;
			}
		}
		if (b != NULL) {
			uint8 flags;
			bool error = false;
#ifndef FRODO_SC
			long vicptr;	// File offset of VIC data
#endif

			while (c != 10)
				c = fgetc(f);	// Shouldn't be necessary
			if (fgetc(f) != 0) {
				ShowRequester("Unknown snapshot format", "OK", NULL);
				fclose(f);
				return false;
			}
			flags = fgetc(f);
#ifndef FRODO_SC
			vicptr = ftell(f);
#endif

			error |= !LoadVICState(f);
			error |= !LoadSIDState(f);
			error |= !LoadCIAState(f);
			error |= !LoadCPUState(f);

			delay = fgetc(f);	// Number of cycles the 6510 is ahead of the previous chips
#ifdef FRODO_SC
			// Make the other chips "catch up" with the 6510
			for (i=0; i<delay; i++) {
				TheVIC->EmulateCycle();
				TheCIA1->EmulateCycle();
				TheCIA2->EmulateCycle();
			}
#endif
			if ((flags & SNAPSHOT_1541) != 0) {
				Prefs *prefs = new Prefs(ThePrefs);
	
				// First switch on emulation
				error |= (fread(prefs->DrivePath[0], 256, 1, f) != 1);
				prefs->Emul1541Proc = true;
				NewPrefs(prefs);
				ThePrefs = *prefs;
				delete prefs;
	
				// Then read the context
				error |= !Load1541State(f);
	
				delay = fgetc(f);	// Number of cycles the 6502 is ahead of the previous chips
#ifdef FRODO_SC
				// Make the other chips "catch up" with the 6502
				for (i=0; i<delay; i++) {
					TheVIC->EmulateCycle();
					TheCIA1->EmulateCycle();
					TheCIA2->EmulateCycle();
					TheCPU->EmulateCycle();
				}
#endif
				Load1541JobState(f);
#ifdef __riscos__
				TheWIMP->ThePrefsToWindow();
#endif
			} else if (ThePrefs.Emul1541Proc) {	// No emulation in snapshot, but currently active?
				Prefs *prefs = new Prefs(ThePrefs);
				prefs->Emul1541Proc = false;
				NewPrefs(prefs);
				ThePrefs = *prefs;
				delete prefs;
#ifdef __riscos__
				TheWIMP->ThePrefsToWindow();
#endif
			}

#ifndef FRODO_SC
			fseek(f, vicptr, SEEK_SET);
			LoadVICState(f);	// Load VIC data twice in SL (is REALLY necessary sometimes!)
#endif
			fclose(f);
	
			if (error) {
				ShowRequester("Error reading snapshot file", "OK", NULL);
				Reset();
				return false;
			} else
				return true;
		} else {
			fclose(f);
			ShowRequester("Not a Frodo snapshot file", "OK", NULL);
			return false;
		}
	} else {
		ShowRequester("Can't open snapshot file", "OK", NULL);
		return false;
	}
}
#endif


#ifdef __BEOS__
#include "C64_Be.i"
#endif

#ifdef AMIGA
#include "C64_Amiga.i"
#endif

#ifdef __unix
#include "C64_x.i"
#endif

#ifdef __mac__
#include "C64_mac.i"
#endif

#ifdef WIN32
#include "C64_WIN32.i"
#endif

#ifdef __riscos__
#include "C64_Acorn.i"
#endif

#ifdef __palmos__
#include "C64_PalmOS.i"
#endif
