/*
 *  C64_Amiga.i - Put the pieces together, Amiga specific stuff
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 */


/*
 *  Constructor, system-dependent things
 */

void C64::c64_ctor1(void)
{
	//ShowRequester(REQ_DEBUG, 0xc7041, 1, 0, 0);
	RAM1 = armletData->ram1;
	RAM2 = armletData->ram2;
        Basic = armletData->basicROM;
        Kernal = armletData->kernalROM;
        Char = armletData->charROM;
        Color = armletData->color;
        RAM1541 = armletData->ram1541;
        ROM1541 = armletData->rom1541;
	//ShowRequester(REQ_DEBUG, 0xc7041, 2, 0, 0);
	if (armletData->cmd & CMD_GETPREFS)
		ThePrefs.GetState(armletData->prefs);
	if (armletData->cmd & CMD_SETPREFS) {
		TheNewPrefs.SetState(armletData->prefs);
		armletData->cmd &= ~(CMD_SETPREFS);
	}
	//ShowRequester(REQ_DEBUG, 0xc7041, 3, 0, 0);
}

void C64::c64_ctor2(void)
{
/*
	if (!TheJob1541 || !TheREU || !TheIEC || !TheCIA2 || !TheCIA1 ||
            !TheSID || !TheVIC || !TheCPU1541 || !TheCPU || !TheDisplay) {
		strcpy(armletData->errMsg, "Some of the chips could not be allocated!");
		armletData->errCode=4;
	}
*/
}


/*
 *  Destructor, system-dependent things
 */

void C64::c64_dtor(void)
{
}


/*
 *  Start emulation
 */

void C64::Run(void)
{
	// Reset chips
	//ShowRequester(REQ_DEBUG, 0x6510b007, 0,0,0);
	TheCPU->Reset();
	//ShowRequester(REQ_DEBUG, 0x6581b007, 0,0,0);
	TheSID->Reset();
	//ShowRequester(REQ_DEBUG, 0x6526b007, 1,0,0);
	TheCIA1->Reset();
	//ShowRequester(REQ_DEBUG, 0x6526b007, 2,0,0);
	TheCIA2->Reset();
	//ShowRequester(REQ_DEBUG, 0x6502b007, 0,0,0);
	TheCPU1541->Reset();

	// Patch kernal IEC routines
	orig_kernal_1d84 = Kernal[0x1d84];
	orig_kernal_1d85 = Kernal[0x1d85];
	//ShowRequester(REQ_DEBUG, 0x9a7c8, 0,0,0);
	PatchKernal(ThePrefs.FastReset, ThePrefs.Emul1541Proc);

	// Start the CPU thread
	thread_running = true;
	quit_thyself = false;
	have_a_break = false;

	if (!armletData->errCode) thread_func();
}


/*
 *  Stop emulation
 */

void C64::Quit(void)
{
	// Ask the thread to quit itself if it is running
	if (thread_running) {
		ShowRequester(REQ_DEBUG, 0xbaddead, 1,0,0);
		quit_thyself = true;
		thread_running = false;
	}
}


/*
 *  Pause emulation
 */

void C64::Pause(void)
{
	TheSID->PauseSound();
}


/*
 *  Resume emulation
 */

void C64::Resume(void)
{
	TheSID->ResumeSound();
}


/*
 *  Vertical blank: Poll keyboard and joysticks, update window
 */

void C64::VBlank(bool draw_frame)
{
	long speed_index;

	/* Jump back to 68k code.
	   This handles events, timing, redrawing, sound playback...
	   All results are stuffed into appropriate fields in armletData.
	   This is the only place the armlet calls the 68k program, except
	   ShowRequester, new and delete.
	*/
	//armletData->drawFrame=draw_frame;
	(armletData->call68k)
		(armletData->emulStateP,
		 (unsigned long)armletData->vBlank,
                 &armletData,
		 4);
		 
	// Commands
	//if (armletData->cmd & ~(CMD_DRAWFRAME))
	//	ShowRequester(REQ_DEBUG, 0xDEB6C0DE, armletData->cmd, 0, 0);
	if (armletData->cmd & CMD_GETPREFS)
		ThePrefs.GetState(armletData->prefs);
	if (armletData->cmd & CMD_SETPREFS) {
		TheNewPrefs.SetState(armletData->prefs);
		NewPrefs(&TheNewPrefs);
		ThePrefs = TheNewPrefs;
		armletData->cmd &= ~(CMD_SETPREFS);
	}
	if (armletData->cmd & CMD_LOAD) {
		LoadSnapshot(armletData->snapshot1, armletData->snapshot2);
		armletData->cmd &= ~(CMD_LOAD);
	}
	if (armletData->cmd & CMD_SAVE)
		SaveSnapshot(armletData->snapshot1, armletData->snapshot2);
	if (armletData->cmd & CMD_RESET) {
		Reset();
		armletData->cmd &= ~(CMD_RESET);
	}

	// Debug, show PC in GUI
	TheCPU->PollPC();
	// Poll keyboard
	TheDisplay->PollKeyboard(TheCIA1->KeyMatrix, TheCIA1->RevMatrix, &joykey);

	// Poll joysticks
	TheCIA1->Joystick1 = poll_joystick(0);
	TheCIA1->Joystick2 = poll_joystick(1);

	if (ThePrefs.JoystickSwap) {
		uint8 tmp = TheCIA1->Joystick1;
		TheCIA1->Joystick1 = TheCIA1->Joystick2;
		TheCIA1->Joystick2 = tmp;
	}

	// Joystick keyboard emulation
	if (TheDisplay->NumLock())
		TheCIA1->Joystick1 &= joykey;
	else
		TheCIA1->Joystick2 &= joykey;

	// Count TOD clocks
	TheCIA1->CountTOD();
	TheCIA2->CountTOD();

	// Update window if needed
	/*
	if (draw_frame) {
		// The 68k callback already handled this
		// TheDisplay->Update();

		// Calculate time between VBlanks, display speedometer
		end_time=armletData->time;
		speed_index = 20 * 100 * ThePrefs.SkipFrames / (end_time-start_time);
		TheDisplay->Speedometer(speed_index);
	}
	*/
}


/*
 *  Open/close joystick drivers given old and new state of
 *  joystick preferences
 */

void C64::open_close_joysticks(bool oldjoy1, bool oldjoy2, bool newjoy1, bool newjoy2)
{
/*
	if (game_open && (oldjoy2 != newjoy2))

		if (newjoy2) {	// Open joystick
			joy_state = 0xff;
			port_allocated = false;

			// Allocate game port
			BYTE ctype;
			Forbid();
			game_io->io_Command = GPD_ASKCTYPE;
			game_io->io_Data = &ctype;
			game_io->io_Length = 1;
			DoIO((struct IORequest *)game_io);

			if (ctype != GPCT_NOCONTROLLER)
				Permit();
			else {
				ctype = GPCT_ABSJOYSTICK;
				game_io->io_Command = GPD_SETCTYPE;
				game_io->io_Data = &ctype;
				game_io->io_Length = 1;
				DoIO((struct IORequest *)game_io);
				Permit();

				port_allocated = true;

				// Set trigger conditions
				game_trigger.gpt_Keys = GPTF_UPKEYS | GPTF_DOWNKEYS;
				game_trigger.gpt_Timeout = 65535;
				game_trigger.gpt_XDelta = 1;
				game_trigger.gpt_YDelta = 1;
				game_io->io_Command = GPD_SETTRIGGER;
				game_io->io_Data = &game_trigger;
				game_io->io_Length = sizeof(struct GamePortTrigger);
				DoIO((struct IORequest *)game_io);

				// Flush device buffer
				game_io->io_Command = CMD_CLEAR;
				DoIO((struct IORequest *)game_io);

				// Start reading joystick events
				game_io->io_Command = GPD_READEVENT;
				game_io->io_Data = &game_event;
				game_io->io_Length = sizeof(struct InputEvent);
				SendIO((struct IORequest *)game_io);
			}

		} else {	// Close joystick

			// Abort game_io
			if (!CheckIO((struct IORequest *)game_io)) {
				AbortIO((struct IORequest *)game_io);
				WaitIO((struct IORequest *)game_io);
			}

			// Free game port
			if (port_allocated) {
				BYTE ctype = GPCT_NOCONTROLLER;
				game_io->io_Command = GPD_SETCTYPE;
				game_io->io_Data = &ctype;
				game_io->io_Length = 1;
				DoIO((struct IORequest *)game_io);

				port_allocated = false;
			}
		}
*/
}

/*
 *  Poll joystick port, return CIA mask
 */

uint8 C64::poll_joystick(int port)
{
	uint8 joy_state=0xff;
	
	if (port==0 && !armletData->prefs->Joystick1On)
		return 0xff;
	if (port==1 && !armletData->prefs->Joystick2On)
		return 0xff;
		
	if (armletData->joystick & JOYFIRE)
		joy_state &= 0xef;
	if (armletData->joystick & JOYRIGHT)
		joy_state &= 0xf7;
	if (armletData->joystick & JOYLEFT)
		joy_state &= 0xfb;
	if (armletData->joystick & JOYDOWN)
		joy_state &= 0xfd;
	if (armletData->joystick & JOYUP)
		joy_state &= 0xfe;

	return joy_state;
}


/*
 * The emulation's main loop
 */
	
void C64::thread_func(void)
{
	while (!quit_thyself) {

#ifdef FRODO_SC
		// The order of calls is important here
		if (TheVIC->EmulateCycle())
			TheSID->EmulateLine();
		TheCIA1->CheckIRQs();
		TheCIA2->CheckIRQs();
		TheCIA1->EmulateCycle();
		TheCIA2->EmulateCycle();
		TheCPU->EmulateCycle();

		if (ThePrefs.Emul1541Proc) {
			TheCPU1541->CountVIATimers(1);
			if (!TheCPU1541->Idle)
				TheCPU1541->EmulateCycle();
		}
		CycleCounter++;
#else
		// The order of calls is important here
		//ShowRequester(REQ_DEBUG, 0x6569d017, 0,0,0);
		int cycles = TheVIC->EmulateLine();
		//ShowRequester(REQ_DEBUG, 0x6581d017, 0,0,0);
		TheSID->EmulateLine();
#if !PRECISE_CIA_CYCLES
		//ShowRequester(REQ_DEBUG, 0x6526d017, 1,0,0);
		TheCIA1->EmulateLine(ThePrefs.CIACycles);
		//ShowRequester(REQ_DEBUG, 0x6526d017, 2,0,0);
		TheCIA2->EmulateLine(ThePrefs.CIACycles);
#endif

		if (ThePrefs.Emul1541Proc) {
			int cycles_1541 = ThePrefs.FloppyCycles;
			TheCPU1541->CountVIATimers(cycles_1541);

			if (!TheCPU1541->Idle) {
				// 1541 processor active, alternately execute
				//  6502 and 6510 instructions until both have
				//  used up their cycles
				while (cycles >= 0 || cycles_1541 >= 0)
					if (cycles > cycles_1541)
						cycles -= TheCPU->EmulateLine(1);
					else
						cycles_1541 -= TheCPU1541->EmulateLine(1);
			} else
				TheCPU->EmulateLine(cycles);
		} else {
			// 1541 processor disabled, only emulate 6510
			//ShowRequester(REQ_DEBUG, 0x6510d017, 0,0,0);
			TheCPU->EmulateLine(cycles);
		}
#endif
		if (armletData->errCode)
			quit_thyself=1;
	}
	if (armletData->errCode==0)
		ShowRequester(REQ_DEBUG, 0xbaddead, 2,0,0);
}

uint8 C64::Random(void)
{
        seed = seed * 1103515245L + 12345;
        return (seed >> 16)&0xff;
}
void C64::SeedRandom(uint32 s)
{
	seed=s;
}

