/*
 *  Display_PalmOS.i - C64 graphics display, emulator window handling,
 *                     PalmOS specific stuff
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 */

#include "C64.h"
#include "SAM.h"
#include "Version.h"


/*
 *  Display constructor: Create window/screen
 */

C64Display::C64Display(C64 *the_c64) : TheC64(the_c64)
{
	// Everything has already been set up by the 68k front end
}


/*
 *  Display destructor
 */

C64Display::~C64Display()
{
}

/*
 *  Prefs may have changed
 */

void C64Display::NewPrefs(Prefs *prefs)
{
}


/*
 *  Redraw bitmap
 */

void C64Display::Update(void)
{
	// Maybe redraws of the leds...
}

/*
 *  Update speedometer
 */

void C64Display::Speedometer(int speed)
{
	// Handled by the 68k part
/*
	static int delay = 0;

	if (delay >= 20) {
		char str[16];
		sprintf(str, "%d%%", speed);
		SetAPen(the_rp, pens[15]);	// Light gray
		RectFill(the_rp, xo+1, yo+DISPLAY_Y+1, xo+DISPLAY_X/5-2, yo+DISPLAY_Y+14);
		SetAPen(the_rp, pens[0]);	// Black
		SetFont(the_rp, speedo_font);
		Move(the_rp, xo+24, yo+DISPLAY_Y+10);
		Text(the_rp, str, strlen(str));
		delay = 0;
	} else
		delay++;
*/
}


/*
 *  Return pointer to bitmap data
 */

uint8 *C64Display::BitmapBase(void)
{
	return TheC64->armletData->displayBuf;
}


/*
 *  Return number of bytes per row
 */

int C64Display::BitmapXMod(void)
{
	return TheC64->armletData->displayModulo;
}


/*
 *  Handle IDCMP messages
 */

void C64Display::PollKeyboard(uint8 *key_matrix, uint8 *rev_matrix, uint8 *joystick)
{
	// The key in armletData is simply the key number left to right, top to bottom
	// Convert to C64 row/column
	
	uint32 direction = (TheC64->armletData->key>>16)&0xff;
	int8 c64_byte = (TheC64->armletData->key>>8)&0xff;
	int8 c64_bit = TheC64->armletData->key&0xff;

/*
	if (TheC64->armletData->key!=0 && TheC64->armletData->key!=0xffff)
		ShowRequester(REQ_DEBUG, 0xfeed, c64_byte, c64_bit, direction);
*/
	if (c64_byte != -1) {
		// Normal keys
		bool shifted = c64_byte & 8;
		c64_byte &= 7;
		if (direction == KEYDOWN) {
		
			// Key pressed
			if (shifted) {
				key_matrix[6] &= 0xef;
				rev_matrix[4] &= 0xbf;
			}
			key_matrix[c64_byte] &= ~(1 << c64_bit);
			rev_matrix[c64_bit] &= ~(1 << c64_byte);
		} else if (direction == KEYUP) {
			// Key released
			if (shifted) {
				key_matrix[6] |= 0x10;
				rev_matrix[4] |= 0x40;
			}
			key_matrix[c64_byte] |= (1 << c64_bit);
			rev_matrix[c64_bit] |= (1 << c64_byte);
		}
	}
}


/*
 *  Check if NumLock is down (for switching the joystick keyboard emulation)
 */

bool C64Display::NumLock(void)
{
	return 0;
}


/*
 *  Allocate C64 colors
 */

void C64Display::InitColors(uint8 *colors)
{
	// Spread pens into colors array, no-brainer in Palm version
	for (int i=0; i<256; i++)
		colors[i] = i&0xf; //pens[i & 0x0f];
}


/*
 *  Show a requester
 */

int32 ShowRequester(int32 id, int32 arg1, int32 arg2, int32 arg3, int32 arg4)
{
	Int32 args[5]={id, arg1, arg2, arg3, arg4};
	Int32 *argp=&(args[0]);
	ArmletData *ad=(ArmletData *)GetR6();
	UInt32 retval;

	PutR6(ad->r6org);
	retval=ad->call68k(ad->emulStateP, (uint32)ad->requester, &argp, 4);
	ad->r6org=GetR6();
	PutR6((uint32)ad);

	return retval;
}
