/*
 *  PrefsState.h - Global preferences, as a struct
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 */

#ifndef _PREFSSTATE_H
#define _PREFSSTATE_H

// The options used in the Palm version, as a struct.
// This struct needs to be aligned exactly the same way with the ARM and the
// 68k. I use int32 exclusively to be on the safe side.

// Drive types
enum {
        DRVTYPE_DIR,    // 1541 emulation in host file system
        DRVTYPE_D64,    // 1541 emulation in .d64 file
        DRVTYPE_T64     // 1541 emulation in .t64 file
};

#define JOYPEN_GRAFFITI 1
#define JOYPEN_SCREEN   2

// Pixels on screen:	12345678
#define SCREENMODE_16		0L
// Pixels in int32:	87654321
#define SCREENMODE_16ALT	1L
// Pixels in int32:	78563412
#define SCREENMODE_256		2L
// Pixels in int32:	04030201 08070605
#define SCREENMODE_256_BUFFERED	3L
#define is256(x) ((x)==SCREENMODE_256 || (x)==SCREENMODE_256_BUFFERED)

struct PrefsState {
	int32 NormalCycles;		// CPU cycles in normal raster lines
	int32 BadLineCycles;		// CPU cycles in Bad Lines
	int32 CIACycles;		// CIA timer ticks per raster line
	int32 FloppyCycles;		// 1541 CPU cycles per line

	int32 AdaptiveFrameSkip;	// Skip frames as nessecary
	int32 SkipFrames;		// Draw every n-th frame
	int32 LimitSpeed;		// Limit speed to 100%

	int32 DriveType[4];		// Type of drive 8..11

	char DrivePath[4][32];		// "Path" for drive 8..11

	int32 SIDType;			// SID emulation type
	int32 REUSize;			// Size of REU

	int32 SpritesOn;		// Sprite display is on
	int32 SpriteCollisions;		// Sprite collision detection is on
	int32 Joystick1On;		// Joystick connected to port 1 of host
	int32 Joystick2On;		// Joystick connected to port 2 of host
	int32 JoystickSwap;		// Swap joysticks 1<->2
	int32 FastReset;		// Skip RAM test on reset
	int32 CIAIRQHack;		// Write to CIA ICR clears IRQ
	int32 Emul1541Proc;		// Enable processor-level 1541 emulation
	int32 SIDFilters;		// Emulate SID filters

// PalmOS only prefs beneath this line

	int32 ScreenMode;
	int32 BordersOn;
	int32 Volume;			// Sound volume (1024==100%)
	int32 JoyPen;			// JoyPen mode
	int32 KeyUp;			// Key for up
	int32 KeyLeft;			// Key for left
	int32 KeyRight;			// Key for right
	int32 KeyDown;			// Key for down
	int32 KeyFire;			// Key for fire button
};
#endif
